import logging
from decorator import decorator

try:
    from web3.datastructures import AttributeDict
except ModuleNotFoundError:
    from web3.utils.datastructures import AttributeDict

def add_0x(value: str) -> str:
    if value.startswith("0x"):
        return str(value)
    return "0x" + str(value)


def remove_0x(value: str) -> str:
    if value.startswith("0x"):
        return str(value)[2:]
    return str(value)


def remove_enclosing(value: str, enclosing: str="\"") -> str:
    strval = str(value)
    if strval.startswith(enclosing) and strval.endswith(enclosing):
        return strval[1:-1]
    return strval


def get_default_logger(name) -> logging.RootLogger:
    if not name:
        name = __name__
    logger = logging.getLogger(name)
    handler = logging.StreamHandler()
    formatter = logging.Formatter(
        '%(asctime)s %(name)-12s %(levelname)-8s %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel(logging.INFO)
    return logger

@decorator
def response_to_dict(f, *args, **kwargs):
    resp = f(*args, **kwargs)
    
    if not resp.text:
        return {}

    return resp.json()

@decorator
def response_to_str(f, *args, **kwargs):
    resp = f(*args, **kwargs)

    if not resp.text:
        return ""

    # this removes closing dquotes
    return resp.json()

@decorator
def response_to_attrdict(f, *args, **kwargs):
    resp = f(*args, **kwargs)

    if not resp.text:
        return AttributeDict({})

    # this removes closing dquotes
    return AttributeDict(resp.json())


# -*- coding: utf-8 -*-
"""
.. module:: session
   :synopsis: Contains abstraction of Parity's secretstore sessions: https://wiki.parity.io/Secret-Store

.. moduleauthor:: Adam Nagy <adam.nagy@energyweb.org>

"""

import logging
from typing import Union

import requests
from requests.exceptions import RequestException
try:
    from web3.datastructures import AttributeDict
except ModuleNotFoundError:
    from web3.utils.datastructures import AttributeDict

from .utils import(
    remove_0x,
    get_default_logger,
    response_to_attrdict,
    response_to_str
)

class SecretStoreSessionError(RequestException):
    """ Thrown when an error has occured during the Secret Store session. 

    """

class Session(object):
    """
    The class holding together the secretstore session calls.

    Args:
        ss_endpoint_uri (str): The endpoint where Secret Store is listening for requests (for sessions).
        logger (:py:obj:`logging.Logger`, optional): The logger object. 
            Defaults to None and instantiates a default logger in this case with log level *INFO*.

    Attributes:
        ss_endpoint_uri (str): The endpoint where Secret Store is listening for requests (for sessions).
        logger (:py:obj:`logging.Logger`): The logger object. 

    Raises:
        ValueError: if the secret store's url is not given
    """

    __SEP = "/"

    def __init__(self, ss_endpoint_uri: str, logger: logging.Logger=None):

        self.logger = logger if logger is not None else get_default_logger(__name__)

        if not ss_endpoint_uri:
            self.logger.error("Secret Store endpoint was not given, please pass it as a string, e.g.: \"http://127.0.0.1:8090\"")
            raise ValueError("Secret store endpoint not specified.")

        self.ss_endpoint_uri = str(ss_endpoint_uri)
        if self.ss_endpoint_uri.endswith(Session.__SEP):
            self.ss_endpoint_uri = self.ss_endpoint_uri[:-1]

    def __query(self, f, *args, **kwargs):
        verbose = kwargs.pop("verbose", True)
        try:
            response = f(*args, **kwargs)
        except RequestException as re:
            if verbose:
                self.logger.error(re)
            raise

        if response.status_code != 200:
            re = SecretStoreSessionError(response=response)
            if verbose:
                self.logger.error("Secret store session error. Status code: {}".format(response.status_code))
                self.logger.error("Reason: {}".format(response.reason))
                self.logger.error("Json: {}".format(response.content))
                self.logger.error("Url: {}".format(response.url))
            raise re

        return response

    def __post(self, *args, **kwargs):
        return self.__query(requests.post, *args, **kwargs)

    def __get(self, *args, **kwargs):
        return self.__query(requests.get, *args, **kwargs)

    def __urlbuild(self, *args):
        return Session.__SEP.join(args)

    @response_to_str
    def generateServerKey(self, server_key_id: str, signed_server_key_id: str, threshold: Union[str, int], verbose=True) -> str:
        """Generates server keys.

        Args:
            server_key_id (str): The server key ID.
            signed_server_key_id (str): The server key ID signed by the SS user.
            threshold (str or int): Key threshold value. 
                Please consider the `guidelines <https://wiki.parity.io/>`_ when choosing this value.
            verbose (bool): Whether to log errors. Default: True.

        Returns:
            str: The hex-encoded public portion of the server key.

        Raises:
            :class:`SecretStoreSessionError`, :class:`requests.RequestException`

        """

        return self.__post(self.__urlbuild(self.ss_endpoint_uri,
                                           "shadow",
                                           remove_0x(server_key_id),
                                           remove_0x(signed_server_key_id),
                                           str(threshold)),
                           verbose=verbose)

    @response_to_str
    def generateServerAndDocumentKey(self, server_key_id, signed_server_key_id, threshold, verbose=True) -> str:
        """Generating document key by one of the participating nodes.
        
        While it is possible (and more secure, if you’re not trusting the Secret Store nodes) 
        to run separate server key generation and document key storing sessions, 
        you can generate both keys simultaneously. 

        Args:
            server_key_id (str): The server key ID.
            signed_server_key_id (str): The server key ID signed by the SS user.
            threshold (str or int): Key threshold value. 
                Please consider the `guidelines <https://wiki.parity.io/>`_ when choosing this value.
            verbose (bool): Whether to log errors. Default: True.

        Returns:
            str: The hex-encoded document key, encrypted with requester's public key (ECIES encryption is used) .

        Raises:
            :class:`SecretStoreSessionError`, :class:`requests.RequestException`

        """
        
        return self.__post(self.__urlbuild(self.ss_endpoint_uri,
                                           remove_0x(server_key_id),
                                           remove_0x(signed_server_key_id),
                                           str(threshold)),
                           verbose=verbose)

    @response_to_attrdict
    def shadowRetrieveDocumentKey(self, server_key_id, signed_server_key_id, verbose=True) -> AttributeDict:
        """This session is a preferable way of retrieving previously generated document key.  

        Args:
            server_key_id (str): The server key ID.
            signed_server_key_id (str): The server key ID signed by the SS user.
            verbose (bool): Whether to log errors. Default: True.

        Returns:
            :class:`web3.datastructures.AttributeDict`: The hex-encoded decrypted_secret, common_point and decrypt_shadows fields.

        Raises:
            :class:`SecretStoreSessionError`, :class:`requests.RequestException`

        """
        
        return self.__get(self.__urlbuild(self.ss_endpoint_uri,
                                          "shadow",
                                          remove_0x(server_key_id),
                                          remove_0x(signed_server_key_id)),
                          verbose=verbose)
    @response_to_str
    def retrieveDocumentKey(self, server_key_id, signed_server_key_id, verbose=True) -> str:
        """ Fetches the document key from the secret store.  

        This is the lighter version of the 
        `document key shadow retrieval <https://wiki.parity.io/Secret-Store#document-key-shadow-retrieval-session>`_ session, 
        which returns final document key (though, encrypted with requester public key) if you have enough trust in 
        the Secret Store nodes. During document key shadow retrieval session, document key is not reconstructed 
        on any node, but it requires Secret Store client either to have an access to Parity RPCs, or 
        to run some EC calculations to decrypt the document key.

        Args:
            server_key_id (str): The server key ID.
            signed_server_key_id (str): The server key ID signed by the SS user.
            verbose (bool): Whether to log errors. Default: True.

        Returns:
            str: The hex-encoded document key, encrypted with requester public key (ECIES encryption is used).

        Raises:
            :class:`SecretStoreSessionError`, :class:`requests.RequestException`

        """
        
        return self.__get(self.__urlbuild(self.ss_endpoint_uri,
                                          remove_0x(server_key_id),
                                          remove_0x(signed_server_key_id)),
                          verbose=verbose)
    
    @response_to_str
    def signSchnorr(self, server_key_id, signed_server_key_id, message_hash, verbose=True) -> str:
        """ Schnorr signing session, for computing Schnorr signature of a given message hash. 

        Args:
            server_key_id (str): The server key ID.
            signed_server_key_id (str): The server key ID signed by the SS user.
            message_hash (str): The 256-bit hash of the message that needs to be signed.
            verbose (bool): Whether to log errors. Default: True.

        Returns:
            str: The hex-encoded Schnorr signature (serialized as c || s), encrypted with requester public key (ECIES encryption is used).

        Raises:
            :class:`SecretStoreSessionError`, :class:`requests.RequestException`

        """
        
        return self.__get(self.__urlbuild(self.ss_endpoint_uri,
                                          "schnorr",
                                          remove_0x(server_key_id),
                                          remove_0x(signed_server_key_id),
                                          remove_0x(message_hash)),
                          verbose=verbose)
    
    @response_to_str
    def signEcdsa(self, server_key_id, signed_server_key_id, message_hash, verbose=True) -> str:
        """ ECDSA signing session, for computing ECDSA signature of a given message hash. 

        Args:
            server_key_id (str): The server key ID.
            signed_server_key_id (str): The server key ID signed by the SS user.
            message_hash (str): The 256-bit hash of the message that needs to be signed.
            verbose (bool): Whether to log errors. Default: True.

        Returns:
            str: The hex-encoded ECDSA signature (serialized as r || s || v ), encrypted with requester public key (ECIES encryption is used).

        Raises:
            :class:`SecretStoreSessionError`, :class:`requests.RequestException`

        """
        
        return self.__get(self.__urlbuild(self.ss_endpoint_uri,
                                          "ecdsa",
                                          remove_0x(server_key_id),
                                          remove_0x(signed_server_key_id),
                                          remove_0x(message_hash)),
                          verbose=verbose)

    @response_to_str
    def storeDocumentKey(self, server_key_id, signed_server_key_id, common_point, encrypted_point, verbose=True) -> str:
        """ Binds an externally-generated document key to a server key. 
        
        Useable after a `server key generation session <https://wiki.parity.io/Secret-Store#server-key-generation-session>`_ . 

        Args:
            server_key_id (str): The server key ID.
            signed_server_key_id (str): The server key ID signed by the SS user.
            common_point (str): The hex-encoded common point portion of encrypted document key.
            encrypted_point (str): The hex-encoded encrypted point portion of encrypted document key.
            verbose (bool): Whether to log errors. Default: True.

        Returns:
            str: Empty string if everything was OK (status code 200).

        Raises:
            :class:`SecretStoreSessionError`, :class:`requests.RequestException`
        """
        
        return self.__post(self.__urlbuild(self.ss_endpoint_uri,
                                           "shadow",
                                           remove_0x(server_key_id),
                                           remove_0x(signed_server_key_id),
                                           remove_0x(common_point),
                                           remove_0x(encrypted_point)),
                           verbose=verbose)

    @response_to_str
    def nodesSetChange(self, node_ids_new_set, signature_old_set, signature_new_set, verbose=True) -> str:
        """ Node set change session.
        
        Requires all added, removed and stable nodes to be online for the duration of the session. 
        Before starting the session, you’ll need to generate two administrator’s signatures: `old set` 
        signature and `new set` signature. To generate these signatures, the Secret Store RPC methods should 
        be used: `serversSetHash` and `signRawHash`. 
        
        Args:
            node_ids_new_set (list(str)): Node IDs of the `new set`.
            signature_old_set (str): ECDSA signature of all online node IDs `keccak(ordered_list(staying + added + removing))`.
            signature_new_set (str): ECDSA signature of node IDs that should stay in the Secret Store after the session ends `keccak(ordered_list(staying + added))`.
            verbose (bool): Whether to log errors. Default: True.

        Returns:
            str: Empty string (probably).

        Raises:
            :class:`SecretStoreSessionError`, :class:`requests.RequestException`
        """

        url = self.__urlbuild(self.ss_endpoint_uri,
                              "admin",
                              "servers_set_change",
                              remove_0x(signature_old_set),
                              remove_0x(signature_new_set))
        
        return self.__post(url, json=node_ids_new_set, verbose=verbose)



# -*- coding: utf-8 -*-
"""
.. module:: module
   :synopsis: Contains abstraction of Parity's secretstore module: https://wiki.parity.io/JSONRPC-secretstore-module

.. moduleauthor:: Adam Nagy <adam.nagy@energyweb.org>

"""

import logging

from web3 import Web3
try:
    from web3.datastructures import AttributeDict
except ModuleNotFoundError:
    from web3.utils.datastructures import AttributeDict

from .session import (
    Session,
    SecretStoreSessionError
)

from .utils import (
    add_0x,
    get_default_logger
)


class SecretStore(object):
    """
    The class holding together the secretstore module API- and session calls.

    Args:
        web3 (:class:`web3.Web3`, optional): The :class:`web3.Web3` instance. Defaults to None, in which case
            it tries to `auto import <https://web3py.readthedocs.io/en/stable/providers.html#how-automated-detection-works>`_ it.
        ss_endpoint_uri (str, optional): The endpoint where Secret Store is listening for requests (for sessions). Defaults to None.
        logger (:py:obj:`logging.Logger`, optional): The logger object. 
            Defaults to None and instantiates a default logger in this case with log level *INFO*.

    Attributes:
        web3 (:class:`web3.Web3`): The :class:`web3.Web3` instance.
        session (:class:`.Session`): The :class:`.Session` instance.

    Returns:
        str: The signed hash.
    """

    def __init__(self, web3: Web3=None, ss_endpoint_uri: str=None, logger: logging.Logger=None):

        self.logger = logger if logger is not None else get_default_logger(
            __name__)

        if not ss_endpoint_uri:
            self.logger.warning("SecretStore URL was not given, you will not be able to call sessions. "
                + "Didn't you forget to pass `ss_endpoint_uri`?")
        else:
            self.session = Session(ss_endpoint_uri, logger)

        if web3 is None:
            self.logger.warning(
                "SecretStore: Web3 instance was not given, using \"auto import\" instead.")
            from web3.auto import w3
            self.web3 = w3
        else:
            self.web3 = web3

    def signRawHash(self, account: str, pwd: str, rawhash: str) -> str:
        """Computes recoverable ECDSA signatures.

        Typically used for signatures of server key id and signatures of nodes-set hash
        in the Secret Store.

        Args:
            account (str): The account of SS user.
            pwd (str): The password of SS user.
            rawhash (str): A 256-bit hash to be signed, e.g.: server key id or nodes-set hash.

        Returns:
            str: The signed hash.

        Raises:
            ValueError: If there is an error in the response.
        """

        return self.web3.manager.request_blocking(
            "secretstore_signRawHash",
            [Web3.toChecksumAddress(account), pwd,
             add_0x(rawhash)]
        )

    def encrypt(self, account, pwd, encrypted_key, hex_document) -> str:
        """ You can use it to encrypt a small document.

        An encryption key is needed, typically obtained from the store by running a 
        `document key retrieval session <https://wiki.parity.io/Secret-Store#document-key-retrieval-session>`_ 
        or a `server- and document key generation session <https://wiki.parity.io/Secret-Store#server-and-document-key-generation-session>`_.

        Args:
            account (str): The account of SS user.
            pwd (str): The password of SS user.
            encrypted_key (str): Document key encrypted with requester's public key.
            hex_document (str): Hex encoded document data.

        Returns:
            str: The encrypted secret document.

        Raises:
            ValueError: If there is an error in the response.
        """

        return self.web3.manager.request_blocking(
            "secretstore_encrypt",
            [Web3.toChecksumAddress(account), pwd,
             add_0x(encrypted_key),
             add_0x(hex_document)]
        )

    def decrypt(self, account: str, pwd: str, encrypted_key: str, encrypted_document: str) -> str:
        """This method can be used to decrypt document, encrypted by :func:`encrypt` method before.

        Args:
            account (str): The account of SS user.
            pwd (str): The password of SS user.
            encrypted_key (str): Document key encrypted with requester's public key.
            encrypted_document (str): Encrypted document data, returned by :func:`encrypt`

        Returns:
            str: The decrypted secret document.

        Raises:
            ValueError: If there is an error in the response.
        """

        return self.web3.manager.request_blocking(
            "secretstore_decrypt",
            [Web3.toChecksumAddress(account), pwd,
             add_0x(encrypted_key),
             add_0x(encrypted_document)]
        )

    def generateDocumentKey(self, account: str, pwd: str, server_key: str) -> AttributeDict:
        """Securely generates document key, so that it remains unknown to all key servers.

        Args:
            account (str): The account of SS user.
            pwd (str): The password of SS user.
            server_key (str): The server key, returned by a server key generating session.
        
        Returns:
            :class:`web3.datastructures.AttributeDict`: The document key.
        
        Raises:
            ValueError: If there is an error in the response.
        """

        return self.web3.manager.request_blocking(
            "secretstore_generateDocumentKey",
            [Web3.toChecksumAddress(account), pwd,
             add_0x(server_key)]
        )

    def shadowDecrypt(self, account, pwd, decrypted_secret, common_point, decrypt_shadows, encrypted_document) -> str:
        """This method can be used to decrypt document, encrypted by :func:`encrypt` method before. .
        
        Document key can be obtained by a 
        `document key shadow retrieval session <https://wiki.parity.io/Secret-Store#document-key-shadow-retrieval-session>`_

        Args:
            account (str): The account of SS user.
            pwd (str): The password of SS user.
            decrypted_secret (str): The hex-encoded decrypted secret portion of an encrypted document key.
            common_point (str): The hex-encoded common point portion of an encrypted document key.
            decrypt_shadows (str): The hex-encoded encrypted point portion of an encrypted document key.
            encrypted_document (str): Encrypted document data, returned by :func:`encrypt`.
        
        Returns:
            str: The decrypted secret document.
        
        Raises:
            ValueError: If there is an error in the response.
        """
        return self.web3.manager.request_blocking(
            "secretstore_shadowDecrypt",
            [Web3.toChecksumAddress(account), pwd,
             add_0x(decrypted_secret),
             add_0x(common_point),
             decrypt_shadows,
             add_0x(encrypted_document)]
        )

    def serversSetHash(self, node_ids: list) -> str:
        """Computes the hash of nodes ids, required to compute nodes set signature
         for manual `nodes set change session <https://wiki.parity.io/Secret-Store-Configuration#changing-the-configuration-of-a-set-of-servers>`_ . 

        Args:
            node_ids (list(str)): List of node ID's (public keys).

        Returns:
            str: The hash.
        
        Raises:
            ValueError: If there is an error in the response.
        """

        return self.web3.manager.request_blocking(
            "secretstore_serversSetHash",
            [node_ids]
        )



import pkg_resources

from secretstore.module import SecretStore
from secretstore.session import (
    Session,
    SecretStoreSessionError
)

import secretstore.utils as utils

#__version__ = pkg_resources.get_distribution("secretstore").version
#__all__ = ["__version__", "SecretStore", "Session", "SecretStoreSessionError", "utils"]
__all__ = ["SecretStore", "Session", "SecretStoreSessionError", "utils"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import (
    setup
)

extras_require = {
    "dev": [
        "bumpversion",
        "setuptools>=36.2.0",
        "wheel",
    ],
    "docs": [
        "Sphinx",
        "sphinx-rtd-theme"
    ],
}

extras_require["dev"] = (
    extras_require["docs"] +
    extras_require["dev"]
)

def readme():
    with open('README.md') as f:
        return f.read()

setup(
    name="secretstore",
    version="0.0.1",
    author="Adam Nagy",
    author_email="adam.nagy@energyweb.org",
    description="Python package for Parity's Secret Store API calls and sessions.",
    long_description=readme(),
    long_description_content_type="text/markdown",
    url="https://github.com/energywebfoundation/secretstore-py",
    project_urls={
        "Documentation": "https://secretstore.readthedocs.io/"
    },
    python_requires='>=3.5.3,<4',
    extras_require=extras_require,
    include_package_data=True,
    install_requires=[
        "web3>=4.0.0",
        "requests>=2.16.0,<3.0.0",
        "decorator",
    ],
    license="MIT",
    zip_safe=False,
    keywords='secretstore secret store parity api energyweb ewf',
    packages=["secretstore",],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
